#!/usr/bin/env python

from GuiPlot import *
from GuiUtils import _logerr, _logwarn, _loginfo
from copy import deepcopy, copy
import time, datetime
import logging
import random
import urllib,urllib2
import socket
import re

class FilelightWorkspace(object):
  sessiondef = {'filelight.view':'site'}
  
  class FilelightView(object):
    viewdef = {}
    def __init__(self,workspace):
      self.workspace = workspace
    def _sessionState(self,session):
      result = {'kind':self.pluginname, 'version':0}
      result.update(self._stateHook(session))
      return repr(result)
  
  class FilelightSiteView(FilelightView):
    label='site'
    pluginname='FilelightSite'
    name='Site'
    
    viewdef={
      'group':0,
      'height':1200,
      'width':1200,
      'type':'png',
      'site':'null',
    }
    
    def _stateHook(self,session):
      return {
        'group':session['filelight.site.group'],
        'height':session['filelight.site.height'],
        'width':session['filelight.site.width'],
        'type':session['filelight.site.type'],
        'site':session['filelight.site.site'],
        'sites':self.workspace.source.nodeNames()
      }
      
    def validate(self,key,value):
      return value
      
  
  class FilelightGroupView(FilelightView):
    label='group'
    pluginname='FilelightGroup'
    name='Group'
    
    viewdef={
      'height':800,
      'width':800,
      'type':'png',
      'group':'null'
    }
    
    def _stateHook(self,session):
      return {
        'height':session['filelight.group.height'],
        'width':session['filelight.group.width'],
        'type':session['filelight.group.type'],
        'group':session['filelight.group.group'],
        'groups':self.workspace.source.groupNames()
      }
      
    def validate(self,key,value):
      return value
  
  def __init__(self,gui,rank,category,name,*args):
    _loginfo("FILELIGHT_W: Workspace created")
    gui._addJSFragment("%s/gui-overview.js" % gui.contentpath)
    gui._addJSFragment("%s/gui-filelight.js" % gui.contentpath)
    self.gui = gui
    self.name = name
    self.rank = rank
    self.category = category
    self.views = [self.FilelightSiteView(self),self.FilelightGroupView(self)]
    self.source = None
    for s in gui.sources:
      if s.plothook=='filelight':
        self.source = s
    if self.source==None:
      _logerr("FILELIGHT_W: Couldn't find FilelightSource")
      raise Exception, "FILELIGHT_W Couldn't find source"
    
  def initialiseSession(self,session):
    for key,value in self.sessiondef.iteritems():
      if key not in session:
        session[key]=deepcopy(value)
    viewstr = session['filelight.view']
    view = self._getView(viewstr)
    for key,value in view.viewdef.iteritems():
      if key not in session:
        session['filelight.%s.%s'%(viewstr,key)]=deepcopy(value)
    
  def start(self,session,**kwargs):
    pass
      
  def sessionView(self,session,*args,**kwargs):
    name = kwargs.get('name',None)
    if not isinstance(name,str):
      raise HTTPError(500,"Invalid name parameter")
    view = self._getView(name)
    for key,value in view.viewdef.iteritems():
      if key not in session:
        session['filelight.%s.%s'%(name,key)]=deepcopy(value)
    session['filelight.view'] = view.label
    self.gui._saveSession(session)
    return self._state(session)
      
  def _getView(self,label):
    view = self.views[0]
    for v in self.views:
      if v.label == label:
        view = v
        break
    return view
    
  def _state(self,session):
    view = self._getView(session['filelight.view'])
    self.gui._noResponseCaching()
    state = ("([{kind: 'AutoUpdate', interval: %d, stamp: %d},"
          + "{kind: 'OverviewHeader', workspace: '%s', workspaces: [%s],"
          + " view: '%s', views: [%s]},"
          + "%s])") \
          % (300,
           self.gui.stamp,
           self.name,
           ",".join("{title: '%s', label: '%s'}"
                % (x.name, x.name.lower()) for x in self.gui.workspaces),
                view.name,
                ",".join("{title: '%s', label: '%s'}"
                   % (x.name, x.label) for x in self.views),
                   view._sessionState(session)) 
    #print state
    return state
  
  def sessionState(self,session,*args,**kwargs):
    return self._state(session)
    
  def sessionResetWorkspace(self,session,*args,**kwargs):
    self.initialiseSession(session)
    self.gui._saveSession(session)
    return self._state(session)
    
  def sessionOpt(self,session,*args,**kwargs):
    viewstr = session['filelight.view']
    view = self._getView(viewstr)
    for key,value in kwargs.items():
      if key in view.viewdef:
        value = view.validate(key,value)
        session['filelight.%s.%s'%(viewstr,key)]=value
    self.gui._saveSession(session)
    return self._state(session)
    
  
      
class FilelightSource(object):
  plothook = 'filelight'
  types={'pdf':'application/pdf','ps':'application/postscript','eps':'application/postscript','png':'image/png','svg':  'image/svg+xml','json':'text/plain'}
  
  def __init__(self,server,*args):
    _loginfo("FILELIGHT_S: Source created.")
    self.cache = [e for e in server.extensions if e.exthook=="OverviewCache"][0]
    socket.setdefaulttimeout(60)
    
    self.nodeNames()
    self.groupNames()
  
  def plot(self,*path,**attrs):
    if len(path)==2:
      what = path[0]
      which = path[1]
      
      if not 'type' in attrs:
        attrs['type']='png'
      elif not attrs['type'] in self.types:
        attrs['type']='png'
      
      if not 'width' in attrs:
        attrs['width']=1000
      elif not attrs['width'].isdigit():
        attrs['width']=1000
      else:
        attrs['width']=int(attrs['width'])
        
      if not 'height' in attrs:
        attrs['height']=1000
      elif not attrs['height'].isdigit():
        attrs['height']=1000
      else:
        attrs['height']=int(attrs['height'])
      
      if what=='site' and which in self.nodeNames():
        if 'group' in attrs:
          group = 1
        else:
          group = 0
        return self.types[attrs['type']],self.cache.lockedCacheFetch('FL_NODEPLOT:%s:%s:%sx%s:%s'%(which,attrs['type'],attrs['width'],attrs['height'],group),self._makeSitePlot,which,attrs)
      if what=='group' and which in self.groupNames():
        return self.types[attrs['type']],self.cache.lockedCacheFetch('FL_GROUPPLOT:%s:%s:%sx%s'%(which,attrs['type'],attrs['width'],attrs['height']),self._makeGroupPlot,which,attrs)
        
    return None,None
    
  def prepareSession(self,session):
    pass
  
  def _get_named_dict(self,name,dictlist):
    for d in dictlist:
      if d['name']==name:
        return d
    return None
    
  def _makeSitePlot(self,site,attrs):
    nodedata = self.cache.lockedCacheFetch('FL_NODE:%s'%site,self._nodeData,site)
    
    sizes = {}
    groups = {}
    for block in nodedata['phedex']['block']:
      name = block['name'].split('#')[0]
      size = int(block['bytes'])
      if name in sizes:
        sizes[name]+=size
      else:
        sizes[name]=size
      if len(block['replica'])==1:
        groups[name]=str(block['replica'][0]['group'])
    
    def dataset_name_splitter(name):
      match = re.match(r'/(.*?)/([A-Za-z0-9]+)(.*?)/(.*)',name)
      if match:
        return [match.group(2),match.group(3).lstrip('-_'),match.group(4),match.group(1)]
      return [name]
    
    def dataset_group_name_splitter(name):
      match = re.match(r'/(.*?)/([A-Za-z0-9]+)(.*?)/(.*)',name)
      if match:
        return [str(groups.get(name,'')),match.group(2),match.group(3).lstrip('-_'),match.group(4),match.group(1)]
      return [str(groups.get(name,'')),name]
    
    if 'group' in attrs:
      splitter = dataset_group_name_splitter
    else:
      splitter = dataset_name_splitter
    
        
    baodata = {'name':'ROOT','value':0,'children':[]}
    for n,s in sizes.items():
      parent = baodata
      parent['value'] += s
      for nn in splitter(n):
        child = self._get_named_dict(nn,parent['children'])
        if child:
          child['value']+=s
          parent = child
        else:
          parent['children'].append({'name':nn,'value':s,'children':[]})
          parent = self._get_named_dict(nn,parent['children'])
          
    plot = BaobabChart(baodata,None,{
      'imageType':attrs['type'],
      'height':attrs['height'],
      'width':attrs['width'],
      'title':'%s\n%s'%(site,time.strftime('%H:%M %d/%m/%Y',time.gmtime())),
      'dpi':72,
      'unit':'B'
      })
    imgdata = plot.draw().read()
    
    return imgdata,3600
     
    
  def _makeGroupPlot(self,group,attrs):
    groupdata = self.cache.lockedCacheFetch('FL_GROUP:%s'%group,self._groupData,group)
    
    sizes = {}
    for node in groupdata['phedex']['node']:
      name = node['name']
      size = 0
      if len(node['group'])==1:
        size = int(node['group'][0]['node_bytes'])
      sizes[name]=size
    
  
    def site_name_splitter(name):
      match = re.match(r'(T\d)_([A-Z]+)_(\w+)',name)
      if match:
        return [match.group(1),match.group(2),match.group(3)]
      return [name]
      
    baodata = {'name':'ROOT','value':0,'children':[]}
    for n,s in sizes.items():
      parent = baodata
      parent['value'] += s
      for nn in site_name_splitter(n):
        child = self._get_named_dict(nn,parent['children'])
        if child:
          child['value']+=s
          parent = child
        else:
          parent['children'].append({'name':nn,'value':s,'children':[]})
          parent = self._get_named_dict(nn,parent['children'])
    
    plot = BaobabChart(baodata,None,{
      'imageType':attrs['type'],
      'height':attrs['height'],
      'width':attrs['width'],
      'title':'%s\n%s'%(group,time.strftime('%H:%M %d/%m/%Y',time.gmtime())),
      'dpi':72,
      'unit':'B'
      })
    imgdata = plot.draw().read()
    
    return imgdata,3600
    
  
  def nodeNames(self):
    return self.cache.cacheFetch('FL_NODENAMES',self._nodeNames)
  def groupNames(self):
    return self.cache.cacheFetch('FL_GROUPNAMES',self._groupNames)
  
    
  def _fetchJSON(self,url):
    data = urllib2.urlopen(url).read()
    data = data.replace('null','None')
    data = eval(data) #UNSAFE!
    return data
    
  def _nodeNames(self):
    data = self._fetchJSON("http://cmsweb.cern.ch/phedex/datasvc/json/prod/nodes")
    nodes = sorted([n['name'] for n in data['phedex']['node']])
    return nodes,86400
    
  def _groupNames(self):
    data = self._fetchJSON("http://cmsweb.cern.ch/phedex/datasvc/json/prod/groups")
    groups = sorted([g['name'] for g in data['phedex']['group']])
    return groups,86400
    
  def _nodeData(self,node):
    data = self._fetchJSON( "http://cmsweb.cern.ch/phedex/datasvc/json/prod/blockreplicas?node=%s"%node)
    return data,3600
    
  def _groupData(self,group):
    data = self._fetchJSON("http://cmsweb.cern.ch/phedex/datasvc/json/prod/groupusage?group=%s"%group)
    return data,3600
    
  
  
  